
-module(alloc).
-compile([export_all]).

go() -> % make the alloc server
   Pid = spawn(?MODULE,top,[[apple,pear,grape],[]]),
   io:fwrite("server pid: ~p ~n",[Pid]),
   register(svr,Pid)
   .
go(all) -> % also make some client processes
   go(), % spawn server
   C1 = spawn(?MODULE,clientOne,[]),
   register(c1,C1),
   C2 = spawn(?MODULE,clientTwo,[]),
   register(c2,C2)
   .

clean() -> c1!halt.

top(Free, Allocated) ->
    receive
	{Pid, alloc} ->
            io:fwrite("Server gets Alloc request ~n",[]),
	    top_alloc(Free, Allocated, Pid);
	{Pid ,{release, Resource}} ->
            io:fwrite("Server gets Release ~p~n",[Resource]),
	    Allocated1 = delete({Resource,Pid}, Allocated),
    	    top([Resource|Free], Allocated1);
	halt -> svr_done
    end.

top_alloc([], Allocated, Pid) ->
    Pid ! no,
    top([], Allocated);

top_alloc([Resource|Free], Allocated, Pid) ->
    Pid ! {yes, Resource},
    top(Free, [{Resource,Pid}|Allocated]).



clientOne() ->
   svr ! {self(),alloc},
   receive
      {yes, R} ->
         io:fwrite("(one) gets alloced ~p~n",[R]),
	 timer:sleep(trunc(10000*rand:uniform())),
         io:fwrite("(one) releases ~p~n",[R]),
         svr!{self(),{release,R}},
	 timer:sleep(trunc(10000*rand:uniform())),
         clientOne();
      {no} -> 
         io:fwrite("(one) cannot get a resource",[]),
         clientOne();
      halt -> timer:sleep(4000), c2!halt, c1_done
   end.

clientTwo() ->
   svr ! {self(),alloc},
   svr ! {self(),alloc},
   svr ! {self(),alloc},
   receive
      {yes, R} ->
         io:fwrite("(two) gets alloced ~p~n",[R]),
	 timer:sleep(trunc(10000*rand:uniform())),
         io:fwrite("(two) releases ~p~n",[R]),
         svr!{self(),{release,R}},
	 timer:sleep(trunc(10000*rand:uniform())),
         clientTwo();
      {no} -> 
         io:fwrite("(two) cannot get a resource",[]),
         clientTwo();
      halt -> timer:sleep(4000), svr!halt, c2_done
   end.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% utils for alloc
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

delete(H, [H|T]) -> T;
delete(X, [H|T]) -> [H|delete(X, T)].

lookup(Pid, [{Resource,Pid}|_]) -> Resource;
lookup(Pid, [_|Allocated])      -> lookup(Pid, Allocated).

ran(0) -> done;
ran(N) ->
	R = trunc(10000*rand:uniform()),
	io:fwrite("~p~n",[R]),
	ran(N-1).
